# Candeza

> Extendable TDD test runner written in TypeScript.

Candeza is a asynchronous, extendable TDD test runner build for high performance. It's designed to address what are limitations of
existing testing frameworks, and be as lightweight as possible. It's size is only `35KB`, or `17 KB minified`.

It's syntax are heavily inspired by the most popular test runners - `Tape`, `Mocha`, `AVA` and `Jasmine`. However, there are key differences.
There are no global variabels, and there are no need to manually count and plan - `t.plan()` - assertions for every test.

In fact there is no `t.end()` either because `Candeza` is 100% promise compilant. Errors will be checked automatically, and if there are no
errors, the test will end. Otherwise the test will fail.


**Candeza is still under development.**



## Features

* High performance.
* Runs tests concurrently and serially
* Enforces writing atomic tests
* No implicit globals
* No planning tests or counting assertions
* Built-In assertion library.
* 100% promise support
* Plugins to extend the core
* Hooks
* Clean API

## Suite

Similar to `Mocha` and `Jasmine` you need to create a suite for your tests before you start to test your module.

```js
import Candeza from 'candeza';

const suite = Candeza('My Suite Name');

suite('My first test!', t => {
    t.pass();
});

```

### Delayed suites

If you need to perform asynchronous operations before any of your suites are run, you may delay the suite.
Run Candeza with the --timeout flag. This will delay your suite. You can\t wrap a `setTimeout` around your suites.

```js

// possible #1

const suite = candeza('Delayed suite', { timeout: 5000 });

    suite.test('My Test #1', t => {
      t.pass();
    });

// possible #2

candeza --timeout 5000

// not possible
setTimeout(() => {
    suite.test('My Test #1', t => {
      t.pass();
    });
}, 5000);

```

### Concurrent tests

By default Candeza runs tests concurrently. Concurrency means that the tests must be atomic. They should not depend on
other tests for state.

| Concurrent           | Description                                                                                   |
| -------------------- | -----------------                                                                             |
| `suite`              | Create a new test.                                                                            |
| `suite.only`         | Allows you to run only the specified test within a suite. All other tests will be skipped.    |
| `suite.todo`         | Creates a test placeholder. The output, but will not count toward the total.                  |
| `suite.skip`         | This feature is the inverse of `.only()`, and the test will be skipped when the suite is run. |

#### Examples

```js

// Exclusive test
suite.only('#indexOf()', t => {
    // ...
});

// Inclusive test
suite.skip('#indexOf()', t => {
    // ...
});

// Synchronous test
suite.test('My Test', t => {
    const result = funcReturnsNumber();
    t.equal(result, 2);
});

// Asynchronous test with promises:
suite.test('My Test', t => {
    return funcReturnsPromise()
        .then(result => {
            t.equal(result, 2);
        });
});
```

### Asynchronous tests and promises

All asynchronous tests can return a Promise or use the `done()` callback. If the test returns a Promise, the test
will pass or fail if the promise resolves or rejects, respectively. And if the `done()` callback is used, the test
will fail if the first argument is defined.

If you are using hooks within your suite, the `done()` callback will be the third argument. Otherwise it is the second argument.

```js
// with hooks
suite.test((t, state, done) => {})

// without hooks
suite.test((t, done) => {})
```

#### Example on a asynchronous test with a callback

```js

suite.test('My Test', (t, state, done) => {
    funcWithCallback((err, result) => {
        t.noError(err);
        t.equal(result, 2);
    });
});

suite.test('My Test 2', (t, state, done) => {
    funcWithCallbackNoValue(done);
});
```

Even if an asynchronous mode is used, synchronous errors are caught.

### Serial tests

Similiar to `Mocha` and `Jasmine`, Candeza also supports serial tests (one at a time). Most of the time concurrent test are preferable, however there are times (such as when accessing a database) that
you may want tests to run serially.

Within a suite, serial tests are executed first followed by any concurrent tests. However, multiple suites are run concurrently. So two serial tests
could run at the same time if they are members of different suites.


The following methods are used to create serial tests. There usage is identical to the equivalent concurrent method.

| Serial               | Description                                                                            |
| -------------------- | -----------------                                                                      |
| `suite.serial`       | Same as suite()                                                                        |
| `suite.serial.only`  | Same as suite.only()                                                                   |
| `suite.serial.todo`  | Same as suite.todo()                                                                   |
| `suite.serial.skip`  | Same as suite.skip()                                                                   |

## Hooks

Candeza lets you register hooks that are run before and after your tests, and before and after individual tests.
This allows you to run setup and/or teardown code.

This four hooks works different from other testing fremworks. Mainly beucause they are running serially
to the tests, and contains it's own state. The state can't be transferred to other tests.

| Hooks                | Description                                                                            |
| -------------------- | -----------------                                                                      |
| `before`             | Run before all tests in the suite                                                      |
| `after`              | Run after all tests in the suite.                                                      |
| `beforeEach`         | Run before each individual tests in the suite.                                         |
| `afterEach`          | Run after each individual tests in the suite.                                          |

#### Hook examples without state

```js
suite.before(t => {
    // this runs before all tests
});

suite.after(t => {
    // this runs after all tests
});

suite.beforeEach(t => {
    // this runs before each test
});

suite.afterEach(t => {
    // this runs after each test
});

```

#### Hook examples with state

```js
suite.beforeEach(() => {
    return { // the 'state' needs a return statement
        data: 2
    };
});

suite('Test #1', (t, state) => {
    t.equal(1 + 1, state.data);
    state.date = 3;
});

suite('Test #2', (t, state) => {
    t.equal(2, state.data); // state is still 2, because beforeEach is run individual for each test
});

```

Keep in mind that the beforeEach and afterEach hooks run just before and after a test is run, and that by default tests run concurrently.
None of this hooks can have a title as in `AVA`.

## Timeouts

Timeouts exist only for suite-level and may be applied to entire test “suites”, or disabled via this.timeout(0). This will be
inherited by all nested suites and test-cases that do not override the value.

## Assertions

`Candeza` includes an assertion library that are mixed into the execution object provided
to each test implementation, are are a bit more feature rich than the core assert module.

Contrary to `AVA` - if multiple assertion failures are encountered within a single test all failures will
be displayed.

| Assertion            | Description                                                                             |
| -------------------- | -----------------                                                                       |
| `pass`               | Generate a passing assertion                                                            |
| `fail`               | Assert that a value is false                                                            |
| `true`               | Asserts that a condition is true                                                        |
| `false`              | Asserts that a condition is false.                                                      |
| `truthy`             | An assertion who tests whether a value is truthy, as determined by !!guard.             |
| `falsey`             | Asserts that the target is falsey.                                                      |
| `assert`             | Basic assertion function that is used as the central point of tracking assertions.      |
| `equal`              | Assert that two values are equal                                                        |
| `notEqual`           | Assert that two values are not equal`                                                   |
| `strictEqual`        | Assert that two values are equal with strict comparison                                 |
| `notStrictEqual`     | Assert that two values are not equal with strict comparison                             |
| `deepEqual`          | Assert that two values are deeply equal                                                 |
| `notDeepEqual`       | Assert that two values are not deeply equal                                             |
| `greaterThan`        | Assert that a value is greater than another value                                       |
| `greaterThanOrEqual` | Assert that a value is greater than another value                                       |
| `lessThan`           | Assert that a value is less than another value                                          |
| `lessThanOrEqual`    | Assert that a value is less than or equal to another value                              |
| `noError`            | An assertion that error is falsey.                                                      |
| `ifError`            | An assertion that error if value is truthy.                                             |
| `regex`              | Assert that a value matches a regular expression.                                       |
| `notRegex`           | Assert that a value does not match a regular expression                                 |
| `throws`             | Expected to throw an error, or promise rejects with an error.                           |
| `notThrows`          | Assert that function doesn't throw an error or promise resolves                           |
| `count`              | An asynchronous assertion that eventually executes a callback a precise number of times |

## Reporters

Candeza reporters adjust to the terminal window, and always disable ANSI-escape coloring when the stdio streams are not associated with a TTY.

Only one reporter exist by default, but new reporters can be added through plugins.

The reporter API is similar to js-reports, but not compatible because it's API is poorly designed for
concurrent testing, reporting suite level and hook errors.


| Event                | Description                                                                             |
| -------------------- | -----------------                                                                       |
| `runStart`           | Emitted before tests start running.                                                     |
| `testEnd`            | Emitted after a test has completed.                                                     |
| `suiteEnd`           | Emitted after a suite has completed.                                                    |
| `runEnd`             | Emitted when all suites are completed.                                                  |

### Adding a new reporter

There are two different ways to add a new reporter. Either through the CLI or directly within the test itself.

```js

// CLI
cadeza --reporter 'foo'

// directly
const suite = Candeza('My test suite', { reporter: 'foo' });

```

## CLI

Candeza includes a CLI, which can be run ad hoc with:

```js
$ $(npm bin)/candeza
```

By default, the CLI will assume your tests are in `test/**/*.js`.

### Configuration

You can configure Candeza via CLI arguments or as properties of a Candeza object in your package.json. A full example follows:

```js
$ $(npm bin)/candeza \
  --timeout 3000 \
  --fail-fast true \
  "test/**/*.js"
```

The above is equivalent to the following package.json entry:

```js
"candeza": {
  "files": "test/**/*.js",
  "timeout": 3000,
  "fail-fast": "true"
}
```

These CLI options are available:

```js

Usage: candeza [options] [<file|folder|glob> ...]

  Options:

   -h, --help              Output usage information
   -H, --harmony           Enable all Harmony flags in JavaScript tests
   -V, --version           Output the version number
   -f, --fail-fast         Stop after first failure
   -p, --preprocessor      Preprocessor to use (i.e. typescript, coffescript)
   -r, --reporter <name>   Specify the reporter to use
   -s, --serial            Run tests serially
   -t, --timeout <ms>      Set test suite timeout in milliseconds.
   -w, --watch             Re-run tests when tests and source files change

```

### Watch mode

Candeza comes with an intelligent watch mode. It watches for files to change and runs just those tests that are affected.
You can enable watch mode using the `--watch` or `-w` flags.

```js
candeza foo.js --watch
```

## Performance

Candeza is developed for high performance, and one of the fastest TDD test runners. Every part of the code is highly optimized,
and all use of 3rd party dependencies are reduced to a minimal.

Over 1000 tests (*without any delays*) can be invoked within 0.17s.

It's recommended to use the `fail-fast` option if you have many tests and suites. The tests will then stop after first failure.

## Contributon

You are welcome to contribute!

